@@ -1,29 +1,69 @@
+import copy
 import json
 import shutil
+import subprocess
+import time
+from collections.abc import Mapping
 from pathlib import Path, PosixPath
 from tempfile import TemporaryDirectory
 from typing import Any, Dict, List, Optional, Union
-import subprocess
-import time
-import requests
-from collections.abc import Mapping
 
-import mlflow
-import copy
-import pandas as pd
-import yaml
 import jsonschema
+import mlflow
+import pandas as pd
+import requests
+import yaml
 
 from pynavio.utils import ExampleRequestType, make_env
 from pynavio.utils.json_encoder import JSONEncoder
-from pynavio.utils.schemas import PREDICTION_SCHEMA, \
-    METADATA_SCHEMA, REQUEST_SCHEMA_SCHEMA
+from pynavio.utils.schemas import (METADATA_SCHEMA, PREDICTION_SCHEMA,
+                                   REQUEST_SCHEMA_SCHEMA,
+                                   not_nested_request_schema)
 
+MODEL_SIZE_LIMIT_IN_BYTES = 1000_000_000
 EXAMPLE_REQUEST = 'example_request'
 ARTIFACTS = 'artifacts'
 ArtifactsType = Optional[Dict[str, str]]
 ERROR_KEYS = {'error_code', 'message', 'stack_trace'}
 PREDICTION_KEY = 'prediction'
+pynavio_model_validation = '(pynavio model validation)'
+
+METADATA = 'metadata'
+OOD_DETECTION = 'oodDetection'
+MLMODEL = 'MLmodel'
+REQUEST_SCHEMA = 'request_schema'
+DATASET = 'dataset'
+EXPLANATIONS = 'explanations'
+
+
+def _is_ood_set_to_default_in_metadata(metadata: dict) -> bool:
+    return metadata.get(OOD_DETECTION, '') == 'default'
+
+
+def _is_explanation_set_to_default_in_metadata(metadata: dict) -> bool:
+    return metadata.get(EXPLANATIONS, '') == 'default'
+
+
+def _is_data_provided_in_metadata(metadata: dict) -> bool:
+    return DATASET in metadata
+
+
+def _is_default_ood_enabled_in_metadata(metadata: dict) -> bool:
+    return _is_ood_set_to_default_in_metadata(metadata) and \
+           _is_data_provided_in_metadata(metadata)
+
+
+def _is_default_explanation_enabled_in_metadata(metadata: dict) -> bool:
+    return _is_explanation_set_to_default_in_metadata(metadata) and \
+           _is_data_provided_in_metadata(metadata)
+
+
+def check_zip_size(model_zip, model_size_in_bytes):
+    if Path(model_zip).stat().st_size > model_size_in_bytes:
+        print(f"Warning: the default model.zip size limit is"
+              f" {model_size_in_bytes} bytes. Please reduce the"
+              f" size or contact craftworks support team to"
+              f" increase the default size")
 
 
 def _get_field(yml: dict, path: str) -> Optional[Any]:
@@ -167,12 +207,9 @@ def _read_metadata(model_path: str) -> dict:
     example_request = _read_example_request(model_path, yml)
 
     return {
-        'dataset':
-            pd.read_csv(data_path) if data_path is not None else None,
-        'explanation_format':
-            _get_field(yml, 'metadata.explanations.format'),
-        'example_request':
-            example_request
+        'dataset': pd.read_csv(data_path) if data_path is not None else None,
+        'explanation_format': _get_field(yml, 'metadata.explanations.format'),
+        'example_request': example_request
     }
 
 
@@ -211,25 +248,106 @@ def _get_example_request_df(model_path):
     return pd.DataFrame(data['data'], columns=data['columns'])
 
 
-class _ModelValidator:
+def _validate_schema(data, schema, name='', raise_exception=True):
+    """
+    Validate the given data against the specified JSON schema.
+
+    @param data: The data to validate.
+    @param schema: The jsonschema schema to validate against.
+    @param name: A descriptive name for the data being validated,
+        used in error message
+    @param raise_exception: Whether to raise an exception if the
+        validation fails. If True (default), the function raises a
+        jsonschema ValidationError exception with a
+        descriptive error message.
+    @return: True if the validation passes, False otherwise.
+     Only returned if raise_exception is False.
+    """
+    is_valid = True
+    try:
+        jsonschema.validate(data, schema)
+    except jsonschema.exceptions.ValidationError:
+        if raise_exception:
+            print(f"Error: {pynavio_model_validation} "
+                  f"Error during {name} validation")
+            raise
+        else:
+            is_valid = False
+    return is_valid
+
+
+def is_input_nested(example_request, not_nested_schema):
+    is_not_nested = _validate_schema(example_request,
+                                     not_nested_schema,
+                                     '',
+                                     raise_exception=False)
+    return not is_not_nested
+
+
+def _is_wrapped_by_prediction_call(func):
+    return getattr(func, '__wrapped_by_prediction_call__', False)
+
+
+def _is_model_predict_wrapped_by_prediction_call(model):
+    return _is_wrapped_by_prediction_call(
+        model._model_impl.python_model.predict)
+
+
+class ModelValidator:
+    """
+    A utility class for validating navio mlflow models.
+    Raises jsonschema.exceptions.ValidationError and AssertionError if
+    there are errors in validation.
+    Example usage:
+        >>> ModelValidator()(model_path = '/path/to/my/model')
+
+    """
+
     @staticmethod
     def validate_metadata(model_path):
+        """
+        Validate the metadata: example request file and MLmodel file
+        of the given navio mlflow model.
+
+        @param model_path: The directory path of the model to validate.
+        @raises jsonschema.exceptions.ValidationError:
+        If there is an error during the validation process.
+        @return: None
+        """
         config = _read_mlmodel_yaml(model_path)
-        try:
-            jsonschema.validate(config.get('metadata'), METADATA_SCHEMA)
-        except jsonschema.exceptions.ValidationError:
-            print("Error during MLmodel validation")
-            raise
+        metadata = config.get('metadata')
+        _validate_schema(metadata, METADATA_SCHEMA, "MLmodel")
 
         example_request = _read_example_request(model_path, config)
-        try:
-            jsonschema.validate(example_request, REQUEST_SCHEMA_SCHEMA)
-        except jsonschema.exceptions.ValidationError:
-            print("Error during example request validation")
-            raise
+        _validate_schema(example_request, REQUEST_SCHEMA_SCHEMA,
+                         "example request")
+        if is_input_nested(example_request, not_nested_request_schema()):
+            print('Warning: {pynavio_model_validation} the nested'
+                  ' model input is not supported'
+                  ' by frontend rendering, it will only be possible'
+                  ' to see the example request as plain json in the'
+                  ' try-out or deployment views. Consider using'
+                  ' string representation of nested example in the'
+                  ' example request json.')
+            if _is_default_ood_enabled_in_metadata(metadata) or \
+                    _is_default_explanation_enabled_in_metadata(metadata):
+                print("Warning: {pynavio_model_validation} default"
+                      " ood and explanations"
+                      " are not supported for nested model inputs.")
 
     @staticmethod
     def run_model_io(model_path, model_input=None, **kwargs):
+        """
+        Run the given navio mlflow model file with the given input.
+
+        @param model_path: The directory path of the model to run.
+        @param model_input: optional, the input data for the model's
+        predict method. If None, the example request specified in
+        the model metadata will be used as input.
+        @param kwargs: Additional keyword arguments.
+
+        @return: The input data and the prediction output.
+        """
         model = mlflow.pyfunc.load_model(model_path)
         if model_input is None:
             model_input = _get_example_request_df(model_path)
@@ -237,61 +355,61 @@ class _ModelValidator:
 
     @staticmethod
     def _check_if_prediction_call_is_used(model_path):
-        # In order to check if the pynavio.prediction_call
-        # is used to decorate the predict method of the model,
-        # a corrupt input is provided. If there is an exception
-        # or the expected keys are not present in the model output,
-        # then that means that the decorator is not applied
-
         model = mlflow.pyfunc.load_model(model_path)
-        corrupt_model_input = pd.DataFrame(
-            {'corrupt_model_input_123': [1, 2, 3]})
-
-        import logging
-        logger = logging.getLogger('gunicorn.error')
-        current_loglevel = logger.level
-        logger.setLevel(logging.CRITICAL)  # this is done to prevent
-        # logging the exception traceback from prediction call
-        # as it will be confusing for the user
-
-        error_msg = \
-            "Please use pynavio.prediction_call to decorate " \
-            "the predict method of the model, which will add the " \
-            "needed error keys for error case"
+        used = True  # do not print warning if not sure
+        # try checking the original model's predict function
         try:
-            model_output = model.predict(corrupt_model_input)
-        except Exception:
-            raise KeyError(error_msg)
-        finally:
-            logger.setLevel(current_loglevel)
+            used = _is_model_predict_wrapped_by_prediction_call(model)
+        except AttributeError:
+            pass
+        if not used:
+            print(f"Warning: {pynavio_model_validation} Please use"
+                  f" pynavio.prediction_call to decorate"
+                  f" the predict method of the model, which will add the"
+                  f" needed error keys({ERROR_KEYS}) for error"
+                  f" case to see descriptive"
+                  f" errors from navio for ease of debugging.")
 
-        assert isinstance(model_output, Mapping), "Model " \
-            "output has to be a dictionary"
-
-        if PREDICTION_KEY not in model_output:  # precaution
-            # to allow for models that always return prediction
-
-            assert set(model_output.keys()) == ERROR_KEYS, error_msg
+    @staticmethod
+    def check_zip_size(model_zip, model_size_in_bytes):
+        if Path(model_zip).stat().st_size > model_size_in_bytes:
+            print(f"Warning: {pynavio_model_validation} "
+                  f"the default model.zip size limit is"
+                  f" {model_size_in_bytes} bytes. Please reduce the"
+                  f" size or contact craftworks support team to"
+                  f" increase the default size")
 
     @staticmethod
     def verify_model_output(model_output, **kwargs):
+        """
+        Verify the output of the navio mlflow model.
+
+        @param model_output: The output of the model.
+        @param kwargs: Additional keyword arguments.
+
+        @raises AssertionError: If the output is not valid.
+        @return: None
+        """
+
         def _validate_prediction_schema(model_prediction):
             try:
                 jsonschema.validate(model_prediction, PREDICTION_SCHEMA)
             except jsonschema.exceptions.ValidationError:
-                print(f"Error: The value of model_output['{PREDICTION_KEY}']"
+                print(f"ERROR: {pynavio_model_validation} The value of "
+                      f"model_output['{PREDICTION_KEY}']"
                       " must be one of the following types "
                       "(cannot be nested or mixed type): "
                       "'array','boolean', 'integer', 'number', 'string'")
                 raise
 
         assert isinstance(model_output, Mapping), "Model " \
-            "output has to be a dictionary"
+            f"ERROR: {pynavio_model_validation} output has to be a dictionary"
 
         if PREDICTION_KEY in model_output:
             _validate_prediction_schema(model_output)
         else:
             assert set(model_output.keys()) == ERROR_KEYS, \
+                f"ERROR: {pynavio_model_validation}" \
                 f"The model output has to contain '{PREDICTION_KEY}'" \
                 f" for prediction" \
                 f" as key for the target, independent of" \
@@ -312,15 +430,43 @@ class _ModelValidator:
                 f"the predict method of the model, which will add the " \
                 f"needed error keys for error case"
 
-    def __call__(self, model_path, **kwargs):
+    def _run(self, model_path, model_zip, model_zip_size_limit, **kwargs):
         self.validate_metadata(model_path)
         model_input, model_output = self.run_model_io(model_path)
         self._check_if_prediction_call_is_used(model_path)
         self.verify_model_output(model_output)
+        self.check_zip_size(model_zip, model_zip_size_limit)
+
+    def __call__(self, model_path, model_zip, model_zip_size_limit, **kwargs):
+        try:
+            self._run(model_path, model_zip, model_zip_size_limit, **kwargs)
+        except (jsonschema.exceptions.ValidationError, AssertionError):
+            print(f'{pynavio_model_validation}: Validation failed. Please fix'
+                  f' the identified issues before uploading the model.'
+                  f'{kwargs.get("append_to_failed_msg", "")}')
+            raise
+        print(f'{pynavio_model_validation}: Validation succeeded.'
+              f'{kwargs.get("append_to_succeeded_msg", "")}')
+
+
+def _is_mlflow2():
+    import mlflow
+    from packaging import version
+    return version.parse(mlflow.__version__) >= version.parse("2.0.0")
+
+
+def _convert_to_mlflow2_format(request_data):
+    dataframe_records = pd.DataFrame.from_records(
+        columns=request_data['columns'],
+        data=request_data['data']).\
+        to_json(orient='records')
+    request_data = {"dataframe_records": json.loads(dataframe_records)}
+    return request_data
 
 
 def check_model_serving(model_path: Union[str, Path],
-                        port=5001, request_bodies=None):
+                        port=5001,
+                        request_bodies=None):
     """
     checks model serving with mlflow. This has limitations, e.g.
     the 'conda.env' setup will not be checked.
@@ -344,6 +490,8 @@ def check_model_serving(model_path: Union[str, Path],
 
     try:
         for data in (request_bodies or _fetch_data(model_path)):
+            if _is_mlflow2():
+                data = _convert_to_mlflow2_format(data)
             response = requests.post(
                 URL,
                 data=json.dumps(data, allow_nan=True),
@@ -357,20 +505,45 @@ def check_model_serving(model_path: Union[str, Path],
         time.sleep(2)
 
 
+def _is_valid_sys_dependency_list(sys_dependencies: List[str]) -> bool:
+    return isinstance(sys_dependencies, List) and \
+        all(isinstance(item, str) for item in sys_dependencies)
+
+
+def _add_sys_dependencies(path: str, sys_dependencies: List[str]) -> None:
+    """
+    Writes the system dependencies to a text file, one per line.
+
+    :param path: Path to save the sys_dependencies.txt file.
+    :param sys_dependencies: List of system dependencies.
+    :return: None
+    """
+    if sys_dependencies is None:
+        return
+
+    assert _is_valid_sys_dependency_list(sys_dependencies)
+
+    file_path = Path(path) / 'sys_dependencies.txt'
+
+    with open(file_path, 'w') as f:
+        f.write("\n".join(sys_dependencies))
+
+
 def to_navio(model: mlflow.pyfunc.PythonModel,
              path,
              example_request: ExampleRequestType = None,
              pip_packages: List[str] = None,
              code_path: Optional[List[Union[str, Path]]] = None,
              conda_packages: List[str] = None,
+             sys_dependencies: List[str] = None,
              conda_channels: List[str] = None,
              conda_env: str = None,
              artifacts: ArtifactsType = None,
              dataset: Optional[dict] = None,
              explanations: Optional[str] = None,
              oodd: Optional[str] = None,
-             num_gpus: Optional[int] = 0
-             ) -> Path:
+             num_gpus: Optional[int] = 0,
+             validate_model: Optional[bool] = True) -> Path:
     """
     create a .zip mlflow model file for navio
     Usage: either pip_packages or conda_env need to be set.
@@ -387,15 +560,20 @@ def to_navio(model: mlflow.pyfunc.PythonModel,
     @param code_path: A list of local filesystem paths to Python file
     dependencies (or directories containing file dependencies)
     @param conda_packages: list of conda packages
+    @param sys_dependencies: list of system library dependencies
     @param conda_channels: list of conda channels
     @param conda_env: the path of a conda.yaml file to use. If specified,
     the values of conda_channels, conda_packages and pip_packages would be
     ignored.
     @param artifacts: If not set, need to set example_request
     @param dataset:
-    @param explanations:
-    @param oodd:
+    @param explanations: expected values are ['disabled', 'default', 'plotly']
+     If not set, 'default' is used
+    @param oodd: expected values are ['disabled', 'default'].
+     If not set, 'default' is used
     @param num_gpus:
+    @param validate_model: if the output model should be validated by
+     ModelValidator. On by default(True), to disable set to False.
 
     Note: Please refer to check_model_serving() method and
     https://navio.craftworks.io/docs/guides/navio-models/model_creation/#3-test-model-serving
@@ -412,8 +590,9 @@ def to_navio(model: mlflow.pyfunc.PythonModel,
                             "dependencies (or directories containing file "
                             "dependencies)), but is not a list")
 
-        if any(Path(code_p).resolve() in Path(path).resolve().parents
-               for code_p in code_path):
+        if any(
+                Path(code_p).resolve() in Path(path).resolve().parents
+                for code_p in code_path):
             raise ValueError("any of 'code_path' argument paths cannot"
                              " be a parent of 'path' argument,"
                              f" please change the 'path': {path} to be"
@@ -447,8 +626,25 @@ def to_navio(model: mlflow.pyfunc.PythonModel,
                       explanations=explanations,
                       oodd=oodd,
                       num_gpus=num_gpus)
+        _add_sys_dependencies(path, sys_dependencies)
+        shutil.make_archive(path, 'zip', path)
+        model_zip = Path(path + '.zip')
 
-    _ModelValidator()(path)
+    if validate_model:
+        msg_kwargs = {
+            'append_to_failed_msg': ' To disable validation set '
+                                    'validate_model to False',
+            'append_to_succeeded_msg': ' Note: Please refer to '
+                                       'check_model_serving()'
+                                       ' method and '
+                                       'https://navio.craftworks.io/'
+                                       'docs/guides/navio-models/'
+                                       'model_creation/'
+                                       '#3-test-model-serving'
+                                       ' for testing the model '
+                                       'serving.',
+        }
+        ModelValidator()(path, model_zip, MODEL_SIZE_LIMIT_IN_BYTES,
+                         **msg_kwargs)
 
-    shutil.make_archive(path, 'zip', path)
-    return Path(path + '.zip')
+    return model_zip
